import asyncio
import json
import logging
import os
import time
import uuid
from importlib import import_module
from itertools import product
from pathlib import Path
from typing import Any

import nest_asyncio
import networkx as nx
from prefect import flow, get_run_logger, task
from prefect.task_runners import ThreadPoolTaskRunner
from prefect_dask.task_runners import DaskTaskRunner

from nxbench.benchmarking.config import AlgorithmConfig, DatasetConfig
from nxbench.benchmarking.utils import (
    add_seeding,
    get_available_backends,
    get_benchmark_config,
    get_machine_info,
    get_python_version,
    memory_tracker,
    process_algorithm_params,
)
from nxbench.data.loader import BenchmarkDataManager
from nxbench.validation.registry import BenchmarkValidator

logger = logging.getLogger("nxbench")

nest_asyncio.apply()

os.environ.setdefault(
    "PREFECT_API_DATABASE_CONNECTION_URL",
    "postgresql+asyncpg://prefect_user:pass@localhost:5432/prefect_db",
)
os.environ.setdefault("PREFECT_ORION_DATABASE_CONNECTION_POOL_SIZE", "5")
os.environ.setdefault("PREFECT_ORION_DATABASE_CONNECTION_MAX_OVERFLOW", "10")
os.environ.setdefault("PREFECT_API_URL", "http://127.0.0.1:4200/api")
os.environ.setdefault("PREFECT_ORION_API_ENABLE_TASK_RUN_DATA_PERSISTENCE", "false")
os.environ.setdefault("MAX_WORKERS", "4")

run_uuid = uuid.uuid4().hex


def load_config() -> dict[str, Any]:
    """Load benchmark configuration dynamically."""
    config = get_benchmark_config()
    return {
        "algorithms": config.algorithms,
        "datasets": config.datasets,
        "env_data": config.env_data,
    }


def setup_cache(
    datasets: list[DatasetConfig],
) -> dict[str, tuple[nx.Graph, dict[str, Any]]]:
    """Load and cache datasets to avoid redundant loading."""
    data_manager = BenchmarkDataManager()
    graphs = {}
    for dataset_config in datasets:
        dataset_name = dataset_config.name
        try:
            graph, metadata = data_manager.load_network_sync(dataset_config)
            graphs[dataset_name] = (graph, metadata)
            logger.debug(
                f"Cached dataset '{dataset_name}' with {graph.number_of_nodes()} "
                f"nodes and {graph.number_of_edges()} edges."
            )
        except Exception:
            logger.exception(f"Failed to load dataset '{dataset_name}'")
    return graphs


@task(name="configure_backend", cache_key_fn=None, persist_result=False)
def configure_backend(original_graph: nx.Graph, backend: str, num_thread: int) -> Any:
    """Prepare the backend for execution."""
    if backend == "networkx":
        graph = original_graph
    elif "parallel" in backend:
        try:
            nxp = import_module("nx_parallel")
            graph = nxp.ParallelGraph(original_graph)
            logger.debug("Configured nx_parallel backend.")
        except ImportError:
            logger.exception("nx-parallel backend not available.")
            raise
    elif "cugraph" in backend:
        try:
            cugraph = import_module("nx_cugraph")
            edge_attr = "weight" if nx.is_weighted(original_graph) else None
            graph = cugraph.from_networkx(original_graph, edge_attrs=edge_attr)
            logger.debug("Configured cugraph backend.")
        except ImportError:
            logger.exception("cugraph backend not available.")
            raise
        except Exception:
            logger.exception("Error converting graph to cugraph format")
            raise
    elif "graphblas" in backend:
        try:
            gb = import_module("graphblas")
            ga = import_module("graphblas_algorithms")
            gb.ss.config["nthreads"] = num_thread
            graph = ga.Graph.from_networkx(original_graph)
            logger.debug("Configured graphblas backend.")
        except ImportError:
            logger.exception("graphblas_algorithms backend not available.")
            raise
        except Exception:
            logger.exception("Error converting graph to graphblas format")
            raise
    else:
        logger.error(f"Unsupported backend: {backend}")
        raise ValueError(f"Unsupported backend: {backend}")

    return graph


@task(name="run_algorithm", cache_key_fn=None, persist_result=False)
def run_algorithm(
    graph: Any, algo_config: AlgorithmConfig, num_thread: int, backend: str
) -> tuple[Any, float, int, str | None]:
    """Run the algorithm on the configured backend"""
    logger = get_run_logger()

    try:
        algo_func = algo_config.get_callable(backend)
    except ImportError as e:
        logger.exception(
            f"Could not get a callable for {algo_config.name} from {backend}."
        )
        return None, 0.0, 0, str(e)

    pos_args, kwargs = process_algorithm_params(algo_config.params)

    kwargs = add_seeding(kwargs, algo_func, algo_config.name)

    error = None
    try:
        original_env = {}
        vars_to_set = [
            "NUM_THREAD",
            "OMP_NUM_THREADS",
            "MKL_NUM_THREADS",
            "OPENBLAS_NUM_THREADS",
        ]
        for var_name in vars_to_set:
            original_env[var_name] = os.environ.get(var_name)
            os.environ[var_name] = str(num_thread)

        with memory_tracker() as mem:
            start_time = time.perf_counter()
            # pass the graph plus the processed pos_args and kwargs
            result = algo_func(graph, *pos_args, **kwargs)
            end_time = time.perf_counter()

        execution_time = end_time - start_time
        peak_memory = mem["peak"]
        logger.debug(f"Algorithm '{algo_config.name}' executed successfully.")

    except Exception as e:
        logger.exception("Algorithm run failed")
        execution_time = time.perf_counter() - start_time
        peak_memory = mem.get("peak", 0)
        result = None
        error = str(e)

    finally:
        # restore environment variables
        for var_name in vars_to_set:
            if original_env[var_name] is None:
                del os.environ[var_name]
            else:
                os.environ[var_name] = original_env[var_name]

    return result, execution_time, peak_memory, error


@task(name="validate_results", cache_key_fn=None, persist_result=False)
def validate_results(
    result: Any, algo_config: AlgorithmConfig, graph: Any
) -> tuple[str, str]:
    logger = get_run_logger()
    validator = BenchmarkValidator()
    try:
        validator.validate_result(result, algo_config.name, graph)
        logger.debug(f"Validation passed for algorithm '{algo_config.name}'.")
    except Exception as e:
        logger.warning(f"Validation warning for '{algo_config.name}'")
        return "warning", str(e)
    return "passed", ""


@task(name="collect_metrics", cache_key_fn=None, persist_result=False)
def collect_metrics(
    execution_time: float,
    execution_time_with_preloading: float,
    peak_memory: int,
    graph: Any,
    algo_config: AlgorithmConfig,
    backend: str,
    dataset_name: str,
    num_thread: int,
    validation_status: str,
    validation_message: str,
    error: str | None = None,
) -> dict[str, Any]:
    logger = get_run_logger()

    if not isinstance(graph, nx.Graph) and hasattr(graph, "to_networkx"):
        graph = graph.to_networkx()

    if error:
        metrics = {
            "execution_time": float("nan"),
            "execution_time_with_preloading": float("nan"),
            "memory_used": float("nan"),
            "num_nodes": (graph.number_of_nodes()),
            "num_edges": (graph.number_of_edges()),
            "algorithm": algo_config.name,
            "backend": backend,
            "dataset": dataset_name,
            "num_thread": num_thread,
            "error": error,
            "validation": validation_status,
            "validation_message": validation_message,
        }
    else:
        metrics = {
            "execution_time": execution_time,
            "execution_time_with_preloading": execution_time_with_preloading,
            "memory_used": peak_memory / (1024 * 1024),  # convert to MB
            "num_nodes": (graph.number_of_nodes()),
            "num_edges": (graph.number_of_edges()),
            "algorithm": algo_config.name,
            "backend": backend,
            "dataset": dataset_name,
            "num_thread": num_thread,
            "validation": validation_status,
            "validation_message": validation_message,
        }

    metrics.update(get_machine_info())
    return metrics


@task(name="teardown_specific", cache_key_fn=None, persist_result=False)
def teardown_specific(backend: str):
    logger = get_run_logger()
    if "parallel" in backend and hasattr(nx.config.backends, "parallel"):
        logger.debug("Tearing down parallel backend configurations.")
        nx.config.backends.parallel.active = False
        nx.config.backends.parallel.n_jobs = 1

    if "cugraph" in backend:
        os.environ["NX_CUGRAPH_AUTOCONFIG"] = "False"
    logger.debug(f"Reset environment variables for backend '{backend}'.")


async def run_single_benchmark(
    backend: str,
    num_thread: int,
    algo_config: AlgorithmConfig,
    dataset_config: DatasetConfig,
    original_graph: nx.Graph,
) -> dict[str, Any] | None:
    logger = get_run_logger()
    logger.info(
        f"Running benchmark for dataset '{dataset_config.name}' with backend "
        f"'{backend}' and {num_thread} threads."
    )

    try:
        preload_start = time.perf_counter()
        graph = configure_backend(original_graph, backend, num_thread)
        preload_time = time.perf_counter() - preload_start

        result, execution_time, peak_memory, error = run_algorithm(
            graph=graph,
            algo_config=algo_config,
            num_thread=num_thread,
            backend=backend,
        )

        execution_time_with_preloading = execution_time + preload_time

        validation_status, validation_message = validate_results(
            result, algo_config, graph
        )
        metrics = collect_metrics(
            execution_time=execution_time,
            execution_time_with_preloading=execution_time_with_preloading,
            peak_memory=peak_memory,
            graph=graph,
            algo_config=algo_config,
            backend=backend,
            dataset_name=dataset_config.name,
            num_thread=num_thread,
            validation_status=validation_status,
            validation_message=validation_message,
            error=error,
        )
    except Exception as e:
        metrics = collect_metrics(
            execution_time=float("nan"),
            execution_time_with_preloading=float("nan"),
            peak_memory=0,
            graph=original_graph,
            algo_config=algo_config,
            backend=backend,
            dataset_name=dataset_config.name,
            num_thread=num_thread,
            validation_status="failed",
            validation_message=str(e),
            error=str(e),
        )
    finally:
        teardown_specific(backend)
        logger.info("Teared down resources after benchmarking.")

    return metrics


@flow(
    name="multiverse_benchmark",
    flow_run_name=f"run_{run_uuid}",
    task_runner=ThreadPoolTaskRunner(max_workers=int(os.getenv("MAX_WORKERS"))),
)
async def benchmark_suite(
    algorithms: list[AlgorithmConfig],
    datasets: list[DatasetConfig],
    backends: list[str],
    threads: list[int],
    graphs: dict[str, tuple[nx.Graph, dict[str, Any]]],
) -> list[dict[str, Any]]:
    """Run the full suite of benchmarks in parallel using asyncio."""
    logger = get_run_logger()
    logger.info("Starting benchmark suite.")

    def create_benchmark_subflow(name_suffix: str, resource_type: str, num_thread: int):
        @flow(
            name="benchmark_subflow",
            flow_run_name=name_suffix,
            task_runner=DaskTaskRunner(
                cluster_kwargs={
                    "n_workers": 1,
                    "resources": {resource_type: 1},
                    "threads_per_worker": num_thread,
                    "processes": False,
                }
            ),
        )
        async def benchmark_subflow(
            backend: str,
            num_thread: int,
            algo_config: AlgorithmConfig,
            dataset_config: DatasetConfig,
            original_graph: nx.Graph,
        ) -> dict[str, Any] | None:
            return await run_single_benchmark(
                backend,
                num_thread,
                algo_config,
                dataset_config,
                original_graph,
            )

        return benchmark_subflow

    tasks = []
    for backend, num_thread, algo_config, dataset_config in product(
        backends, threads, algorithms, datasets
    ):
        dataset_name = dataset_config.name
        if dataset_name not in graphs:
            logger.warning(f"Dataset '{dataset_name}' not cached. Skipping.")
            continue
        original_graph, _ = graphs[dataset_name]
        resource_type = "GPU" if backend == "cugraph" else "process"
        name_suffix = f"{algo_config.name}_{dataset_name}_{backend}_{num_thread}"

        unique_subflow = create_benchmark_subflow(
            name_suffix, resource_type, num_thread
        )

        tasks.append(
            unique_subflow(
                backend=backend,
                num_thread=num_thread,
                algo_config=algo_config,
                dataset_config=dataset_config,
                original_graph=original_graph,
            )
        )

    return await asyncio.gather(*tasks, return_exceptions=True)


async def main_benchmark(
    results_dir: Path = Path("results"),
):
    """Execute benchmarks using Prefect."""
    final_results = []
    timestamp = str(time.strftime("%Y%m%d%H%M%S"))

    try:
        config = load_config()
        algorithms = config["algorithms"]
        datasets = config["datasets"]
        env_data = config["env_data"]

        available_backends = (
            get_available_backends()
        )  # e.g. {"networkx": "3.4.2", "graphblas": "2023.10.0"}

        # parse user-specified constraints for Python versions, backends, threads
        pythons = env_data.get("pythons", ["3.10"])
        backend_configs = env_data.get("backend", {"networkx": ["networkx==3.4.1"]})
        num_threads = env_data.get("num_threads", [1])
        if not isinstance(num_threads, list):
            num_threads = [num_threads]
        num_threads = [int(x) for x in num_threads]

        # check Python version
        actual_python_version = get_python_version()  # e.g. "3.10.12"
        if not any(py_ver in actual_python_version for py_ver in pythons):
            logger.error(
                f"No requested Python version matches the actual interpreter "
                f"({actual_python_version}). Aborting."
            )
            return

        # filter out backends not installed or not matching pinned versions
        chosen_backends = []
        backend_version_map = {}
        for backend, requested_versions in backend_configs.items():
            installed_version = available_backends.get(backend)
            if not installed_version:
                continue

            matched = False
            for req_ver in requested_versions:
                if "==" in req_ver:
                    # e.g. "networkx==3.4.1"
                    _, pinned_version = req_ver.split("==", 1)
                    if pinned_version == installed_version:
                        matched = True
                        break
                else:
                    # if the user didn't pin a version, accept the installed version
                    matched = True
                    break

            if matched:
                chosen_backends.append(backend)
                backend_version_map[backend] = installed_version

        if not chosen_backends:
            logger.error("No valid backends found or matched. Exiting.")
            return
        logger.info(
            f"Chosen backends: {chosen_backends} "
            f"(Installed versions: {backend_version_map})"
        )

        graphs = setup_cache(datasets)

        results = await benchmark_suite(
            algorithms=algorithms,
            datasets=datasets,
            backends=chosen_backends,
            threads=num_threads,
            graphs=graphs,
        )

        for run_result in results:
            if isinstance(run_result, BaseException):
                logger.error("A subflow raised an exception: %s", run_result)
                continue

            if isinstance(run_result, dict):
                run_result["python_version"] = actual_python_version
                bname = run_result.get("backend", "unknown")
                run_result["backend_version"] = backend_version_map.get(
                    bname, "unknown"
                )
                final_results.append(run_result)

    finally:
        results_dir = Path("results")
        results_dir.mkdir(exist_ok=True)
        out_file = results_dir / f"{run_uuid}_{timestamp}.json"

        with out_file.open("w") as f:
            json.dump([r for r in final_results if isinstance(r, dict)], f, indent=4)

        logger.info(f"Benchmark suite results saved to {out_file}")



import json
import logging
from pathlib import Path
from typing import Any

import pandas as pd

from nxbench.benchmarking.config import BenchmarkResult
from nxbench.benchmarking.utils import get_python_version
from nxbench.data.db import BenchmarkDB
from nxbench.data.loader import BenchmarkDataManager

logger = logging.getLogger("nxbench")


class ResultsExporter:
    """Handle loading, processing, and exporting of benchmark results."""

    def __init__(self, results_file: Path):
        """Initialize the results exporter.

        Parameters
        ----------
        results_file : Path
            Path to the benchmark results file (JSON or CSV)
        """
        self.results_file = results_file
        self.data_manager = BenchmarkDataManager()
        self._cached_results: list[BenchmarkResult] | None = None

    def load_results(self) -> list[BenchmarkResult]:
        """Load benchmark results from the workflow outputs (JSON or CSV),
        integrating all known fields into BenchmarkResult and treating unknown
        fields as metadata.
        """
        if self._cached_results is not None:
            return self._cached_results

        results = []

        try:
            if self.results_file.suffix.lower() == ".json":
                with self.results_file.open("r") as f:
                    data = json.load(f)
                if not isinstance(data, list):
                    logger.error(
                        f"Expected a list of results in JSON file, got {type(data)}"
                    )
                    return []

                for entry in data:
                    result = self._create_benchmark_result_from_entry(entry)
                    if result:
                        results.append(result)

            elif self.results_file.suffix.lower() == ".csv":
                df = pd.read_csv(self.results_file)
                for _, row in df.iterrows():
                    entry = row.to_dict()
                    result = self._create_benchmark_result_from_entry(entry)
                    if result:
                        results.append(result)
            else:
                logger.error(f"Unsupported file format: {self.results_file.suffix}")
                return []

        except Exception:
            logger.exception(f"Failed to load results from: {self.results_file}")
            return []

        self._cached_results = results
        logger.info(f"Loaded {len(results)} benchmark results from {self.results_file}")
        return results

    def _create_benchmark_result_from_entry(
        self, entry: dict[str, Any]
    ) -> BenchmarkResult | None:
        try:
            known_fields = {
                "algorithm",
                "dataset",
                "execution_time",
                "execution_time_with_preloading",
                "memory_used",
                "num_nodes",
                "num_edges",
                "is_directed",
                "is_weighted",
                "backend",
                "num_thread",
                "date",
                "validation",
                "validation_message",
                "error",
            }

            algorithm = entry.get("algorithm", "unknown")
            dataset = entry.get("dataset", "unknown")
            backend = entry.get("backend", "unknown")
            execution_time = float(entry.get("execution_time", float("nan")))
            execution_time_with_preloading = float(
                entry.get("execution_time_with_preloading", float("nan"))
            )
            memory_used = float(entry.get("memory_used", float("nan")))
            num_thread = int(entry.get("num_thread", 1))
            num_nodes = int(entry.get("num_nodes", 0))
            num_edges = int(entry.get("num_edges", 0))
            is_directed = bool(entry.get("is_directed", False))
            is_weighted = bool(entry.get("is_weighted", False))
            date = int(entry.get("date", 0))
            validation = entry.get("validation", "unknown")
            validation_message = entry.get("validation_message", "")
            error = entry.get("error")

            metadata = {k: v for k, v in entry.items() if k not in known_fields}

            return BenchmarkResult(
                algorithm=algorithm,
                dataset=dataset,
                execution_time=execution_time,
                execution_time_with_preloading=execution_time_with_preloading,
                memory_used=memory_used,
                num_nodes=num_nodes,
                num_edges=num_edges,
                is_directed=is_directed,
                is_weighted=is_weighted,
                backend=backend,
                num_thread=num_thread,
                date=date,
                metadata=metadata,
                validation=validation,
                validation_message=validation_message,
                error=error,
            )
        except Exception:
            logger.exception("Failed to process result entry.")
            return None

    def to_dataframe(self) -> pd.DataFrame:
        results = self.load_results()
        if not results:
            raise ValueError("No benchmark results found")

        records = []
        all_metadata_keys = set()

        for result in results:
            record = {
                "algorithm": result.algorithm,
                "dataset": result.dataset,
                "backend": result.backend,
                "execution_time": result.execution_time,
                ### ADDED:
                "execution_time_with_preloading": result.execution_time_with_preloading,
                ### END ADDED
                "memory_used": result.memory_used,
                "num_nodes": result.num_nodes,
                "num_edges": result.num_edges,
                "is_directed": result.is_directed,
                "is_weighted": result.is_weighted,
                "num_thread": result.num_thread,
                "date": result.date,
                "validation": result.validation,
                "validation_message": result.validation_message,
                "error": result.error,
                "python_version": get_python_version(),
            }

            for k, v in result.metadata.items():
                record[k] = v
                all_metadata_keys.add(k)

            records.append(record)

        df = pd.DataFrame(records)
        for mk in all_metadata_keys:
            if mk not in df.columns:
                df[mk] = pd.NA

        return df

    def export_results(
        self, output_path: Path, form: str = "csv", if_exists: str = "replace"
    ) -> None:
        """Export benchmark results in specified format (csv, sql, json)."""
        df = self.to_dataframe()

        if form == "csv":
            df.to_csv(output_path, index=False)
            logger.info(f"Exported results to CSV: {output_path}")

        elif form == "sql":
            db = BenchmarkDB(output_path)

            if if_exists == "replace":
                db.delete_results()

            results = self.load_results()
            db.save_results(
                results=results,
                machine_info={},
                python_version=get_python_version(),
            )
            logger.info(f"Exported results to SQL database: {output_path}")

        elif form == "json":
            df.to_json(output_path, orient="records", indent=2)
            logger.info(f"Exported results to JSON: {output_path}")

        else:
            raise ValueError(f"Unsupported export format: {form}")

    def query_results(
        self,
        algorithm: str | None = None,
        backend: str | None = None,
        dataset: str | None = None,
        date_range: tuple[str, str] | None = None,
    ) -> pd.DataFrame:
        """Query benchmark results with optional filtering."""
        df = self.to_dataframe()

        if algorithm:
            df = df[df["algorithm"] == algorithm]
        if backend:
            df = df[df["backend"] == backend]
        if dataset:
            df = df[df["dataset"] == dataset]
        if date_range:
            start_date, end_date = date_range
            df = df[
                (df["date"] >= pd.to_datetime(start_date))
                & (df["date"] <= pd.to_datetime(end_date))
            ]

        return df.sort_values(["algorithm", "dataset", "backend"])



import gc
import importlib
import inspect
import logging
import os
import platform
import random
import sys
import tracemalloc
from contextlib import contextmanager
from importlib.metadata import PackageNotFoundError
from importlib.metadata import version as get_version
from pathlib import Path
from typing import TYPE_CHECKING, Any

import numpy as np
import psutil

from nxbench.benchmarking.config import AlgorithmConfig, BenchmarkConfig, DatasetConfig
from nxbench.benchmarking.constants import ALGORITHM_SUBMODULES

if TYPE_CHECKING:
    from collections.abc import Callable

logger = logging.getLogger("nxbench")

_BENCHMARK_CONFIG: BenchmarkConfig | None = None


def configure_benchmarks(config: BenchmarkConfig | str):
    global _BENCHMARK_CONFIG  # noqa: PLW0603
    if _BENCHMARK_CONFIG is not None:
        raise ValueError("Benchmark configuration already set")
    if isinstance(config, BenchmarkConfig):
        _BENCHMARK_CONFIG = config
    elif isinstance(config, str):
        _BENCHMARK_CONFIG = BenchmarkConfig.from_yaml(config)
    else:
        raise TypeError("Invalid type for configuration")


def get_benchmark_config() -> BenchmarkConfig:
    global _BENCHMARK_CONFIG  # noqa: PLW0603
    if _BENCHMARK_CONFIG is not None:
        return _BENCHMARK_CONFIG

    config_file = os.getenv("NXBENCH_CONFIG_FILE")
    if config_file:
        config_path = Path(config_file)

        if not config_path.is_absolute():
            config_path = (Path.cwd() / config_path).resolve()

        if not config_path.exists():
            raise FileNotFoundError(f"Config file not found: {config_path}")

        logger.debug(f"Resolved config file path: {config_path}")

        _BENCHMARK_CONFIG = BenchmarkConfig.from_yaml(str(config_path))
    else:
        _BENCHMARK_CONFIG = load_default_config()

    return _BENCHMARK_CONFIG


def load_default_config() -> BenchmarkConfig:
    default_algorithms = [
        AlgorithmConfig(
            name="pagerank",
            func="networkx.algorithms.link_analysis.pagerank_alg.pagerank",
            params={"alpha": 0.85},
        ),
    ]
    default_datasets = [
        DatasetConfig(name="08blocks", source="networkrepository"),
        DatasetConfig(name="jazz", source="networkrepository"),
        DatasetConfig(name="karate", source="networkrepository"),
        DatasetConfig(name="enron", source="networkrepository"),
    ]

    env_data = {
        "num_threads": ["1", "4"],
        "backend": {
            "networkx": ["networkx==3.4.2"],
            "graphblas": ["graphblas_algorithms==2023.10.0"],
        },
        "pythons": ["3.10", "3.11"],
    }

    return BenchmarkConfig(
        algorithms=default_algorithms,
        datasets=default_datasets,
        env_data=env_data,
        machine_info={},
    )


def is_nx_cugraph_available():
    try:
        import importlib.util
    except ImportError:
        return False
    else:
        return importlib.util.find_spec("nx_cugraph") is not None


def is_graphblas_available():
    try:
        import importlib.util
    except ImportError:
        return False
    else:
        return importlib.util.find_spec("graphblas_algorithms") is not None


def is_nx_parallel_available():
    try:
        import importlib.util
    except ImportError:
        return False
    else:
        return importlib.util.find_spec("nx_parallel") is not None


def get_python_version() -> str:
    """Get formatted Python version string."""
    version_info = sys.version_info
    return f"{version_info.major}.{version_info.minor}.{version_info.micro}"


def get_available_backends() -> dict[str, str]:
    """Return a dict of available backends and their versions."""
    available = {}

    try:
        networkx = importlib.import_module("networkx")
        nx_version = getattr(networkx, "__version__", None)
        if nx_version is None:
            nx_version = get_version("networkx")
        available["networkx"] = nx_version
    except (ImportError, PackageNotFoundError):
        pass

    if is_nx_cugraph_available():
        try:
            nx_cugraph = importlib.import_module("nx_cugraph")
            cugraph_version = getattr(nx_cugraph, "__version__", None)
            if cugraph_version is None:
                cugraph_version = get_version("nx_cugraph")
            available["cugraph"] = cugraph_version
        except (ImportError, PackageNotFoundError):
            pass

    if is_graphblas_available():
        try:
            graphblas_algorithms = importlib.import_module("graphblas_algorithms")
            gb_version = getattr(graphblas_algorithms, "__version__", None)
            if gb_version is None:
                gb_version = get_version("graphblas_algorithms")
            available["graphblas"] = gb_version
        except (ImportError, PackageNotFoundError):
            pass

    if is_nx_parallel_available():
        try:
            nx_parallel = importlib.import_module("nx_parallel")
            parallel_version = getattr(nx_parallel, "__version__", None)
            if parallel_version is None:
                parallel_version = get_version("nx_parallel")
            available["parallel"] = parallel_version
        except (ImportError, PackageNotFoundError):
            pass

    logger.debug(f"Available backends and versions: {available}")
    return available


class MemorySnapshot:
    """Class to store and diff memory snapshots."""

    def __init__(self, snapshot=None):
        """Initialize with optional tracemalloc snapshot."""
        self.snapshot = snapshot

    def take(self):
        """Take a new snapshot."""
        self.snapshot = tracemalloc.take_snapshot()

    def compare_to(self, other: "MemorySnapshot") -> tuple[int, int]:
        """Compare this snapshot to another and return (current, peak) memory diff in
        bytes.
        """
        if not self.snapshot or not other.snapshot:
            return 0, 0

        stats = self.snapshot.compare_to(other.snapshot, "lineno")
        current = sum(stat.size_diff for stat in stats)
        peak = sum(stat.size for stat in stats)
        return current, peak


@contextmanager
def memory_tracker():
    """Track memory usage of code block.

    Returns dict with 'current' and 'peak' memory usage in bytes.
    Memory usage is measured as the difference between before and after execution.
    """
    gc.collect()
    tracemalloc.start()

    baseline = MemorySnapshot()
    baseline.take()

    mem = {}
    try:
        yield mem
    finally:
        gc.collect()
        end = MemorySnapshot()
        end.take()
        current, peak = end.compare_to(baseline)

        mem["current"] = current
        mem["peak"] = peak

        tracemalloc.stop()


def get_available_algorithms():
    """Get algorithms from specified NetworkX submodules and custom
    algorithms.

    Returns
    -------
    Dict[str, Callable]
        Dictionary of available algorithms.
    """
    nx_algorithm_dict: dict[str, Callable] = {}

    for submodule in ALGORITHM_SUBMODULES:
        spec = importlib.util.find_spec(submodule)
        if spec is None:
            continue
        module = importlib.import_module(submodule)

        for attr_name in dir(module):
            if not attr_name.startswith("_") and not any(
                attr_name.startswith(prefix)
                for prefix in [
                    "is_",
                    "has_",
                    "get_",
                    "set_",
                    "contains_",
                    "write_",
                    "read_",
                    "to_",
                    "from_",
                    "generate_",
                    "make_",
                    "create_",
                    "build_",
                    "delete_",
                    "remove_",
                    "not_implemented",
                    "np_random_state",
                ]
            ):
                try:
                    attr = getattr(module, attr_name)
                except AttributeError:
                    continue
                if inspect.isfunction(attr):
                    if "approximation" in module.__name__:
                        nx_algorithm_dict[f"approximate_{attr_name}"] = attr
                    else:
                        nx_algorithm_dict[attr_name] = attr

    return nx_algorithm_dict


def get_machine_info():
    info = {
        "arch": platform.machine(),
        "cpu": platform.processor(),
        "num_cpu": str(psutil.cpu_count(logical=True)),
        "os": f"{platform.system()} {platform.release()}",
        "ram": str(psutil.virtual_memory().total),
    }
    # info["docker"] = os.path.exists("/.dockerenv")
    return info


def process_algorithm_params(
    params: dict[str, Any],
) -> tuple[list[Any], dict[str, Any]]:
    """Process and separate algorithm parameters into positional and keyword arguments.

    1. Keys starting with "_" go into pos_args (list).
    2. Other keys become kwargs (dict).
    3. If a param is a string that looks like a float or int, parse it.
    4. If param is a dict containing {"func": "..."} then dynamically load that
    function.
    """
    pos_args = []
    kwargs = {}

    for key, value in params.items():
        # attempt to parse string values as float or int:
        if isinstance(value, str):
            try:
                if "." in value or "e" in value.lower():
                    value = float(value)
                else:
                    value = int(value)
            except ValueError:
                pass  # not numeric, leave it as-is

        if isinstance(value, dict) and "func" in value:
            module_path, func_name = value["func"].rsplit(".", 1)
            module = __import__(module_path, fromlist=[func_name])
            value = getattr(module, func_name)

        if key.startswith("_"):
            pos_args.append(value)
        else:
            kwargs[key] = value

    return pos_args, kwargs


def add_seeding(kwargs: dict, algo_func: Any, algorithm_name: str) -> dict:
    # 1. Retrieve optional fields from kwargs
    #    a) user_seed: integer to set global seeds
    #    b) use_local_random_state: boolean (if True, we might pass a local np.random.
    # RandomState)
    user_seed = kwargs.pop("seed", None)
    use_local_random_state = kwargs.pop("use_local_random_state", False)

    # 2. If user_seed is an int, set global seeds
    if isinstance(user_seed, int):
        random.seed(user_seed)
        np.random.seed(user_seed)
        logger.debug(f"Global random seeds set to {user_seed}.")

    # 3. Introspect the function signature to see if it takes 'seed' or 'random_state'
    func_sig = inspect.signature(algo_func)
    can_accept_seed = "seed" in func_sig.parameters
    can_accept_random_state = "random_state" in func_sig.parameters

    # 4. If can_accept_seed and we have an int user_seed, pass it as a kwarg
    if can_accept_seed and isinstance(user_seed, int):
        kwargs["seed"] = user_seed
        logger.debug(
            f"Passing `seed={user_seed}` to algorithm function {algorithm_name}."
        )

    # 5. If can_accept_random_state and user requested a local RandomState
    #    we create one and pass it in
    local_random_state = None
    if can_accept_random_state and use_local_random_state:
        # if user_seed is int, let's seed the local RNG with it, else use default.
        local_random_state = np.random.RandomState(
            user_seed if isinstance(user_seed, int) else None
        )
        kwargs["random_state"] = local_random_state
        logger.debug(
            f"Created local RandomState for algorithm {algorithm_name}, seed="
            f"{user_seed}."
        )
    return kwargs



import logging

import pandas as pd

logger = logging.getLogger("nxbench")


def load_data(file_path="results/results.csv") -> pd.DataFrame:
    """Load the raw CSV data into a Pandas DataFrame."""
    df = pd.read_csv(file_path, dtype=str)
    # hack to handle the deprecated "iteritems" rename in recent Pandas versions
    pd.DataFrame.iteritems = pd.DataFrame.items
    return df


def preprocess_data(df: pd.DataFrame) -> pd.DataFrame:
    """Clean and transform the raw DataFrame"""
    essential_columns = ["algorithm", "execution_time", "memory_used"]
    if "execution_time_with_preloading" not in df.columns:
        df["execution_time_with_preloading"] = df["execution_time"]
    df = df.dropna(subset=essential_columns)

    df["execution_time"] = pd.to_numeric(df["execution_time"], errors="coerce")
    df["execution_time_with_preloading"] = pd.to_numeric(
        df["execution_time_with_preloading"], errors="coerce"
    )
    df["memory_used"] = pd.to_numeric(df["memory_used"], errors="coerce")
    df["num_nodes"] = pd.to_numeric(df["num_nodes"], errors="coerce")
    df["num_edges"] = pd.to_numeric(df["num_edges"], errors="coerce")
    df["num_thread"] = pd.to_numeric(df["num_thread"], errors="coerce")

    df["execution_time_with_preloading"] = df["execution_time_with_preloading"].fillna(
        df["execution_time"]
    )

    df = df.dropna(
        subset=[
            "algorithm",
            "execution_time",
            "execution_time_with_preloading",
            "memory_used",
        ]
    )

    string_columns = [
        "algorithm",
        "dataset",
        "backend",
        "is_directed",
        "is_weighted",
        "python_version",
        "backend_version",
        "cpu",
        "os",
    ]
    for col in string_columns:
        if col in df.columns:
            df[col] = df[col].astype(str).str.strip().str.lower()

    unique_n_nodes = df["num_nodes"].nunique(dropna=True)
    if unique_n_nodes > 1:
        num_nodes_binned = pd.cut(df["num_nodes"], bins=min(unique_n_nodes, 4))

        node_labels = []
        for interval in num_nodes_binned.cat.categories:
            lower = int(interval.left) if pd.notnull(interval.left) else float("-inf")
            upper = int(interval.right) if pd.notnull(interval.right) else float("inf")
            node_labels.append(f"{lower} <= x < {upper}")

        node_label_map = dict(zip(num_nodes_binned.cat.categories, node_labels))
        df["num_nodes_bin"] = num_nodes_binned.replace(node_label_map)
    else:
        df["num_nodes_bin"] = df["num_nodes"]

    unique_n_edges = df["num_edges"].nunique(dropna=True)
    if unique_n_edges > 1:
        num_edges_binned = pd.cut(df["num_edges"], bins=min(unique_n_edges, 4))

        edge_labels = []
        for interval in num_edges_binned.cat.categories:
            lower = int(interval.left) if pd.notnull(interval.left) else float("-inf")
            upper = int(interval.right) if pd.notnull(interval.right) else float("inf")
            edge_labels.append(f"{lower} <= x < {upper}")

        edge_label_map = dict(zip(num_edges_binned.cat.categories, edge_labels))
        df["num_edges_bin"] = num_edges_binned.replace(edge_label_map)
    else:
        df["num_edges_bin"] = df["num_edges"]

    return df


def aggregate_data(df: pd.DataFrame) -> tuple[pd.DataFrame, list, list]:
    """Aggregate the preprocessed DataFrame by computing the mean execution time,
    mean memory usage, etc., grouping by relevant columns.
    """
    group_columns = [
        "algorithm",
        "dataset",
        "backend",
        "num_nodes_bin",
        "num_edges_bin",
        "is_directed",
        "is_weighted",
        "python_version",
        "cpu",
        "os",
        "num_thread",
    ]

    if "backend_version" in df.columns:
        df["backend_version"] = df["backend_version"].apply(
            lambda x: (
                x.split("==")[1] if isinstance(x, str) and "==" in x else "unknown"
            )
        )
        df["backend_full"] = df.apply(
            lambda row: (
                f"{row['backend']} ({row['backend_version']})"
                if row["backend_version"] != "unknown"
                else row["backend"]
            ),
            axis=1,
        )
        group_columns = [
            c if c != "backend" else "backend_full"
            for c in group_columns
            if c != "backend_version"
        ]
    else:
        logger.warning("No 'backend_version' column found in the dataframe.")
        group_columns = [c for c in group_columns if c not in ("backend_version",)]

    df_agg = df.groupby(group_columns, as_index=False, observed=True).agg(
        mean_execution_time=("execution_time", "mean"),
        mean_memory_used=("memory_used", "mean"),
        sample_count=("execution_time", "size"),
        mean_preload_execution_time=("execution_time_with_preloading", "mean"),
    )

    df_agg.set_index(group_columns, inplace=True)

    df_index = df_agg.index.to_frame()
    unique_counts = df_index.nunique()
    available_parcats_columns = [
        col for col in group_columns if col != "algorithm" and unique_counts[col] > 1
    ]

    return df_agg, group_columns, available_parcats_columns


def load_and_prepare_data(file_path: str, logger: logging.Logger):
    """Orchestrate data loading, preprocessing, and aggregation. Returns the final
    aggregated data, along with any metadata needed for downstream visualization.
    """
    raw_df = load_data(file_path)
    cleaned_df = preprocess_data(raw_df)
    df_agg, group_columns, available_parcats_columns = aggregate_data(cleaned_df)
    return cleaned_df, df_agg, group_columns, available_parcats_columns
